"""
Shuffle input sequences within a sliding window, keeping mononuc compo.

Written by Luis del Peso
Modified by A. Mathelier

Modified by Aziz Khan on October 29, 2019
Modified by A. Mathelier in March 2020

"""

from __future__ import print_function
from Bio.SeqRecord import SeqRecord
from Bio.Seq import Seq
from Bio.Alphabet import generic_dna
from biasaway.utils import GC, dinuc_count
import random


def shuffle_window(ss, wl, step):
    """
    Shuffle sequence in the given window.

    Return the shuffled sequence.

    """

    bs = ss[:]
    for i in range(0, len(bs) - 1, step):
        substring = bs[i:(i + wl)]
        bs = ''.join([bs[0:i],
                      ''.join(random.sample(substring, len(substring))),
                      bs[i + wl:]])
    return bs


def generate_sequences(seqs, winlen, step, nfold):
    """
    Shuffle sequences within a sliding window, keeping mononuc compo.

    Return %GC and length distribution of output sequences.

    """

    cpt = 1
    bg_gc_list = []
    bg_lengths = []
    dinuc = [0] * 16
    for record in seqs:
        sequence = record.seq.__str__()
        descr = "Background sequence for {0:s}".format(record.name, cpt)
        for _ in range(0, nfold):
            new_sequence = shuffle_window(sequence, winlen, step)
            new_seq = SeqRecord(Seq(new_sequence, generic_dna),
                                id="background_seq_{0:d}".format(cpt),
                                description=descr)
            print(new_seq.format("fasta"), end='')
            bg_gc_list.append(GC(new_sequence))
            bg_lengths.append(len(new_sequence))
            dinuc = [x + y for x, y in zip(dinuc, dinuc_count(new_sequence))]
            cpt += 1
    return bg_gc_list, bg_lengths, dinuc



# altschulEriksonDinuclShuffle.py
# P. Clote, Oct 2003
# NOTE: One cannot use function "count(s,word)" to count the number
# of occurrences of dinucleotide word in string s, since the built-in
# function counts only nonoverlapping words, presumably in a left to
# right fashion.


import sys,string,random


def computeCountAndLists(s):
  #WARNING: Use of function count(s,'UU') returns 1 on word UUU
  #since it apparently counts only nonoverlapping words UU
  #For this reason, we work with the indices.

  #Initialize lists and mono- and dinucleotide dictionaries
  List = {} #List is a dictionary of lists
  List['A'] = []; List['C'] = [];
  List['G'] = []; List['T'] = [];
  nuclList   = ["A","C","G","T"]
  s       = s.upper()
  s       = s.replace("T","T")
  nuclCnt    = {}  #empty dictionary
  dinuclCnt  = {}  #empty dictionary
  for x in nuclList:
    nuclCnt[x]=0
    dinuclCnt[x]={}
    for y in nuclList:
      dinuclCnt[x][y]=0

  #Compute count and lists
  nuclCnt[s[0]] = 1
  nuclTotal     = 1
  dinuclTotal   = 0
  for i in range(len(s)-1):
    x = s[i]; y = s[i+1]
    List[x].append( y )
    nuclCnt[y] += 1; nuclTotal  += 1
    dinuclCnt[x][y] += 1; dinuclTotal += 1
  assert (nuclTotal==len(s))
  assert (dinuclTotal==len(s)-1)
  return nuclCnt,dinuclCnt,List
 
 
def chooseEdge(x,dinuclCnt):
  numInList = 0
  for y in ['A','C','G','T']:
    numInList += dinuclCnt[x][y]
  z = random.random()
  denom=dinuclCnt[x]['A']+dinuclCnt[x]['C']+dinuclCnt[x]['G']+dinuclCnt[x]['T']
  numerator = dinuclCnt[x]['A']
  if z < float(numerator)/float(denom):
    dinuclCnt[x]['A'] -= 1
    return 'A'
  numerator += dinuclCnt[x]['C']
  if z < float(numerator)/float(denom):
    dinuclCnt[x]['C'] -= 1
    return 'C'
  numerator += dinuclCnt[x]['G']
  if z < float(numerator)/float(denom):
    dinuclCnt[x]['G'] -= 1
    return 'G'
  dinuclCnt[x]['T'] -= 1
  return 'T'


def connectedToLast(edgeList,nuclList,lastCh):
  D = {}
  for x in nuclList: D[x]=0
  for edge in edgeList:
    a = edge[0]; b = edge[1]
    if b==lastCh: D[a]=1
  for i in range(2):
    for edge in edgeList:
      a = edge[0]; b = edge[1]
      if D[b]==1: D[a]=1
  ok = 0
  for x in nuclList:
    if x!=lastCh and D[x]==0: return 0
  return 1
 

def eulerian(s):
  nuclCnt,dinuclCnt,List = computeCountAndLists(s)
  #compute nucleotides appearing in s
  nuclList = []
  for x in ["A","C","G","T"]:
    if x in s: nuclList.append(x)
  #compute numInList[x] = number of dinucleotides beginning with x
  numInList = {}
  for x in nuclList:
    numInList[x]=0
    for y in nuclList:
      numInList[x] += dinuclCnt[x][y]
  #create dinucleotide shuffle L 
  firstCh = s[0]  #start with first letter of s
  lastCh  = s[-1]
  edgeList = []
  for x in nuclList:
    if x!= lastCh: edgeList.append( [x,chooseEdge(x,dinuclCnt)] )
  ok = connectedToLast(edgeList,nuclList,lastCh)
  return ok,edgeList,nuclList,lastCh


def shuffleEdgeList(L):
  n = len(L); barrier = n
  for i in range(n-1):
    z = int(random.random() * barrier)
    tmp = L[z]
    L[z]= L[barrier-1]
    L[barrier-1] = tmp
    barrier -= 1
  return L


def dinuclShuffle(s):
  ok = 0
  while not ok:
    ok,edgeList,nuclList,lastCh = eulerian(s)
  nuclCnt,dinuclCnt,List = computeCountAndLists(s)

  #remove last edges from each vertex list, shuffle, then add back
  #the removed edges at end of vertex lists.
  for [x,y] in edgeList: List[x].remove(y)
  for x in nuclList: shuffleEdgeList(List[x])
  for [x,y] in edgeList: List[x].append(y)

  #construct the eulerian path
  L = [s[0]]; prevCh = s[0]
  for i in range(len(s)-2):
    ch = List[prevCh][0] 
    L.append( ch )
    del List[prevCh][0]
    prevCh = ch
  L.append(s[-1])
  #t = string.join(L,"") for p3k by Aziz
  t = "".join(L)
  return t



"""
Generation of sequences using mono-nucleotide of input sequences.

Module allowing the generation of sequences by using a mono-nucleotide.
shuffling of the given sequences.

Modified by Aziz Khan on October 29, 2019
Modified by A. Mathelier in March 2020

"""

from __future__ import print_function
from Bio.SeqRecord import SeqRecord
from Bio.Seq import Seq
from Bio.Alphabet import generic_dna
from biasaway.utils import GC, dinuc_count
import random


def generate_sequences(seqs, nfold):
    """
    Generate sequences by shuffling input (mononucleotide).

    return tuple containing %GC compo and length distrib of output.

    """

    cpt = 1
    bg_gc_list = []
    bg_lengths = []
    dinuc = [0] * 16
    for record in seqs:
        seq = record.seq.__str__()
        descr = "Background sequence for {0:s}".format(record.name)
        for _ in range(0, nfold):
            new_sequence = ''.join(random.sample(seq, len(seq)))
            new_seq = SeqRecord(Seq(new_sequence, generic_dna),
                                id="background_seq_{0:d}".format(cpt),
                                description=descr)
            print(new_seq.format("fasta"), end='')
            bg_gc_list.append(GC(new_sequence))
            bg_lengths.append(len(new_sequence))
            dinuc = [x + y for x, y in zip(dinuc, dinuc_count(new_sequence))]
            cpt += 1
    return bg_gc_list, bg_lengths, dinuc



# Luis del Peso
# Modified by A. Mathelier
# Vancouver, Jan 2012

# Modified by Aziz Khan on October 29, 2019
# Modified by A. Mathelier in March 2020


from __future__ import print_function
from Bio.SeqRecord import SeqRecord
from Bio.Seq import Seq
from Bio.Alphabet import generic_dna
from biasaway.altschulEriksonDinuclShuffle import dinuclShuffle
from biasaway.utils import GC, dinuc_count, split_seq
import re


def shuffle_window(ss, wl, step):
    bs = ss[:]
    for i in range(0, len(bs)-1, step):
        bs = bs[0:i] + dinuclShuffle(bs[i:(i+wl)]) + bs[i+wl:]
    return(bs)  # returns shuffled sequence


def generate_sequences(seqs, winlen, step, nfold):
    cpt = 1
    bg_gc_list = []
    bg_lengths = []
    dinuc = [0] * 16
    for record in seqs:
        seq = record.seq.__str__()
        descr = "Background sequence for {0:s}".format(record.name, cpt)
        for n in range(0, nfold):
            new_sequence = ""
            for sequence in split_seq(seq):
                if re.match('N', sequence):
                    new_sequence += sequence
                elif sequence:
                    new_sequence += shuffle_window(sequence, winlen, step)
            new_seq = SeqRecord(Seq(new_sequence, generic_dna),
                                id="background_seq_{0:d}".format(cpt),
                                description=descr)
            print(new_seq.format("fasta"), end='')
            bg_gc_list.append(GC(new_sequence))
            bg_lengths.append(len(new_sequence))
            dinuc = [x + y for x, y in zip(dinuc, dinuc_count(new_sequence))]
            cpt += 1
    return bg_gc_list, bg_lengths, dinuc



"""
Module matching %GC compo distribution b/w fg and bg.

Modified by Aziz Khan on October 29, 2019
Modified by A. Mathelier in March 2020

"""


from __future__ import print_function
import sys
import random
from Bio import SeqIO
from biasaway.utils import GC, dinuc_count, open_for_parsing


def fg_GC_bins(fg_file):
    """
    Compute G+C content for all sequences in the foreground.

    It computes the %GC content and store the information in a list. To each
    G+C percentage bin, we associate the number of sequences falling in the
    corresponding bin.
    Return lists of GC contents, GC bins, lengths distrib, and dinuc compo.

    """
    with open_for_parsing(fg_file) as stream:
        gc_bins = [0] * 101
        gc_list = []
        lengths = []
        dinuc = [0] * 16
        for record in SeqIO.parse(stream, "fasta"):
            gc = GC(record.seq)
            gc_list.append(gc)
            # python 3 fix
            gc = int(round(gc))
            gc_bins[gc] += 1
            lengths.append(len(record.seq))
            dinuc = [x + y for x, y in zip(dinuc, dinuc_count(record.seq))]
    return gc_list, gc_bins, lengths, dinuc


def fg_len_GC_bins(fg_file):
    """
    Compute G+C content for all sequences in the foreground.

    Computes %GC contant and store the information in a list. To each G+C
    percentage bin, we associate the number of sequences falling in the
    corresponding bin.
    Return lists of GC contents, GC bins, and lengths distrib.

    """
    with open_for_parsing(fg_file) as stream:
        gc_bins = []
        for _ in range(0, 101):
            gc_bins.append({})
        gc_list = []
        lengths = []
        dinuc = [0] * 16
        for record in SeqIO.parse(stream, "fasta"):
            gc = GC(record.seq)
            gc_list.append(gc)
            length = len(record)
            dinuc = [x + y for x, y in zip(dinuc, dinuc_count(record.seq))]
            lengths.append(length)
            if length in gc_bins[gc]:
                gc_bins[gc][length] += 1
            else:
                gc_bins[gc][length] = 1
    return gc_list, gc_bins, lengths, dinuc


def print_rec(rec, stream):
    """ Print a record to a stream output. """

    stream.write("{0}\n".format(rec.format("fasta")))


def print_in_bg_dir(gc_bins, bg_dir, with_len=False):
    """ Print the sequences in the bg directory in bin files. """

    for percent in range(0, 101):
        with open_for_parsing("{0}/bg_bin_{1}.txt".format(bg_dir, percent), 'w') as stream:
            if with_len:
                for length in gc_bins[percent]:
                    for rec in gc_bins[percent][length]:
                        print_rec(rec, stream)
            else:
                for rec in gc_bins[percent]:
                    print_rec(rec, stream)


def bg_GC_bins(bg_file, bg_dir):
    """
    Compute G+C content for all sequences in the background.

    Compute and store the GC information in a list. To each G+C percentage bin,
    we associate the corresponding sequence names.
    Files representing the binning are stored in the "bg_dir" directory.
    Return lists of GC contents, GC bins, and lengths distrib.

    """
    with open_for_parsing(bg_file) as stream:
        gc_bins = []
        gc_list = []
        lengths = []
        dinuc = [0] * 16
        for _ in range(0, 101):
            gc_bins.append([])
        for record in SeqIO.parse(stream, "fasta"):
            gc = GC(record.seq)
            gc_list.append(gc)
            gc = int(round(gc))  # python3 fix
            gc_bins[gc].append(record)
            lengths.append(len(record.seq))
            dinuc = [x + y for x, y in zip(dinuc, dinuc_count(record.seq))]
    print_in_bg_dir(gc_bins, bg_dir)
    return gc_list, gc_bins, lengths, dinuc


def bg_len_GC_bins(bg_file, bg_dir):
    """
    Compute G+C content for all sequences in the background.

    Compute and store the %GC information in a list. To each G+C percentage
    bin, we associate the corresponding sequence names.
    Return lists of GC contents, GC bins, and lengths distrib.

    """
    with open_for_parsing(bg_file) as stream:
        gc_bins = []
        gc_list = []
        lengths = []
        dinuc = [0] * 16
        for _ in range(0, 101):
            gc_bins.append({})
        for record in SeqIO.parse(stream, "fasta"):
            gc = GC(record.seq)
            gc_list.append(gc)
            if len(record) in gc_bins[gc]:
                gc_bins[gc][len(record)].append(record)
            else:
                gc_bins[gc][len(record)] = [record]
            lengths.append(len(record.seq))
            dinuc = [x + y for x, y in zip(dinuc, dinuc_count(record.seq))]
    print_in_bg_dir(gc_bins, bg_dir, True)
    return gc_list, gc_bins, lengths, dinuc


def get_bins_from_bg_dir(bg_dir, percent):
    """ Return the sequences from the corresponding bin file. """

    with open_for_parsing("{0}/bg_bin_{1:d}.txt".format(bg_dir, percent)) as stream:
        bin_seq = []
        for record in SeqIO.parse(stream, "fasta"):
            bin_seq.append(record)
        return bin_seq


def generate_sequences(fg_bins, bg_bins, bg_dir, nfold):
    """
    Choose randomly the background sequences in each bin of %GC.

    Follow the same distribution as the one of foreground sequences with a
    nfold ratio.
    Return the list of %GC and length distrib.

    """
    lengths = []
    gc_list = []
    dinuc = [0] * 16
    for percent in range(0, 101):
        if fg_bins[percent]:
            random.seed()
            try:
                nb = fg_bins[percent] * nfold
                if bg_bins:
                    bin_seq = bg_bins[percent]
                else:
                    bin_seq = get_bins_from_bg_dir(bg_dir, percent)
                sample = random.sample(bin_seq, nb)
                gc_list.extend([percent] * nb)
            except ValueError:
                sys.stderr.write("""*** WARNING ***
                    Sample larger than population for {0:d}% G+C content:
                    {1:d} needed and {2:d} obtained\n""".format(
                    percent, fg_bins[percent] * nfold, len(bin_seq)))
                sample = bin_seq
                gc_list.extend([percent] * len(bin_seq))
            for r in sample:
                print(r.format("fasta"), end='')
                lengths.append(len(r.seq))
                dinuc = [x + y for x, y in zip(dinuc, dinuc_count(r.seq))]
    return gc_list, lengths, dinuc


def extract_seq_rec(size, nb, bg_keys, bg, accu, index):
    """
    Extract "nb" sequences with sizes equal to "size" nt.

    We try to get exact size or as close as possible to "size" nt.  This is a
    tail recursive function with the accumulator "accu" looking for sizes
    "bg_keys" in the bg set "bg".
    Return the accumulator and the number of found sequences.

    """
    if not (bg_keys and nb):  # End of the recursion since we have no sequence
        # in the bg or enough retrieved (nb=0)
        return accu, nb
    if index > len(bg_keys) - 1:
        return extract_seq_rec(size, nb, bg_keys, bg, accu, index - 1)
    if not bg_keys:  # No more size in the background to be checked so return
        # what was in the previous size bin
        if bg[bg_keys[index - 1]]:
            random.shuffle(bg[bg_keys[index - 1]])
            accu.extend(bg[bg_keys[index - 1]][0:nb])
            bg[bg_keys[index - 1]] = bg[index - 1][nb:]
            return accu, nb - len(bg[bg_keys[index - 1]][0:nb])
        else:
            return accu, nb
    if bg_keys[index] >= size:  # No need to go further in the different sizes
        # within the background
        if (index == 0 or not bg[bg_keys[index - 1]] or
                bg_keys[index] - size < size - bg_keys[index - 1]):  # Which
            # size is the closest to the expected one? => we go for the current
            # one if YES
            random.shuffle(bg[bg_keys[index]])
            accu.extend(bg[bg_keys[index]][0:nb])
            new_nb = nb - len(bg[bg_keys[index]][0:nb])
            if bg[bg_keys[index]][nb:]:  # Check that there is sequences in the
                # background for this size bin
                bg[bg_keys[index]] = bg[bg_keys[index]][nb:]
                return extract_seq_rec(size, new_nb, bg_keys, bg, accu, index)
            else:
                bg[bg_keys[index]] = bg[bg_keys[index]][nb:]
                del bg_keys[index]
                return extract_seq_rec(size, new_nb, bg_keys, bg, accu, index)
        else:  # The previous size was the closest
            random.shuffle(bg[bg_keys[index - 1]])
            accu.extend(bg[bg_keys[index - 1]][0:nb])
            new_nb = nb - len(bg[bg_keys[index - 1]][0:nb])
            if bg[bg_keys[index - 1]][nb:]:  # Check that there is sequences in
                # the background for this size bin
                bg[bg_keys[index - 1]] = bg[bg_keys[index - 1]][nb:]
                return extract_seq_rec(size, new_nb, bg_keys, bg, accu, index)
            else:
                bg[bg_keys[index - 1]] = bg[bg_keys[index - 1]][nb:]
                del bg_keys[index - 1]
                return extract_seq_rec(size, new_nb, bg_keys, bg, accu,
                                       index - 1)
    elif index == len(bg_keys) - 1:
        random.shuffle(bg[bg_keys[index]])
        accu.extend(bg[bg_keys[index]][0:nb])
        new_nb = nb - len(bg[bg_keys[index]][0:nb])
        if bg[bg_keys[index]][nb:]:
            bg[bg_keys[index]] = bg[bg_keys[index]][nb:]
            return extract_seq_rec(size, new_nb, bg_keys, bg, accu, index)
        else:
            bg[bg_keys[index]] = bg[bg_keys[index]][nb:]
            del bg_keys[index]
            return extract_seq_rec(size, new_nb, bg_keys, bg, accu, index)
    else:
        return extract_seq_rec(size, nb, bg_keys, bg, accu, index + 1)


def get_bins_len_from_bg_dir(bg_dir, percent):
    """ Return the sequences from the corresponding bin file. """

    with open_for_parsing("{0}/bg_bin_{1:d}.txt".format(bg_dir, percent)) as stream:
        bin_seq = {}
        for record in SeqIO.parse(stream, "fasta"):
            length = len(record)
            if length in bin_seq:
                bin_seq[length].append(record)
            else:
                bin_seq[length] = [record]
        return bin_seq


def generate_len_sequences(fg, bg, bg_dir, nfold):
    """
    Extract the sequences from the bg with similar sizes as in the fg.

    Return the %GC list and length distrib.

    """

    sys.setrecursionlimit(10000)
    random.seed()
    lengths = []
    gc_list = []
    dinuc = [0] * 16
    for percent in range(0, 101):
        if fg[percent]:
            nb = sum(fg[percent].values()) * nfold
            sequences = []
            for size in fg[percent].keys():
                nb_to_retrieve = fg[percent][size] * nfold
                if bg:
                    bg_bins = bg[percent]
                else:
                    bg_bins = get_bins_len_from_bg_dir(bg_dir, percent)
                bg_keys = sorted(bg_bins.keys())
                seqs, _ = extract_seq_rec(size, nb_to_retrieve, bg_keys,
                                          bg_bins, [], 0)
                sequences.extend(seqs)
            nb_match = len(sequences)
            gc_list.extend([percent] * nb_match)
            if nb_match != nb:
                sys.stderr.write("""*** WARNING ***
                    Sample larger than population for {0:d}% G+C content:
                    {1:d} needed and {2:d} obtained\n""".format(percent,
                                                                nb,
                                                                nb_match))
            for s in sequences:
                lengths.append(len(s))
                dinuc = [x + y for x, y in zip(dinuc, dinuc_count(s.seq))]
                print("{0:s}".format(s.format("fasta")), end='')
    return gc_list, lengths, dinuc



#!/usr/bin/env python

"""
This is a setup script for BiasAway: a tool for DNA sequence background generation

This code is free software; you can redistribute it and/or modify it under the terms of the
BSD License (see the file LICENSE.md included with the distribution).

@author: Aziz Khan
@email: aziz.khan@ncmm.uio.no
"""
import os
from distutils.core import setup
from setuptools import find_packages
from biasaway import __version__ as VERSION


CLASSIFIERS = [
    'Intended Audience :: Developers',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: MIT License',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Topic :: Scientific/Engineering :: Bio-Informatics',
    'Topic :: Software Development :: Libraries :: Python Modules',
]

install_requires = [
    'biopython',
    'numpy',
]

#def readme():
#    with open('README.rst') as f:
#        return f.read()

def readme(fname):
    with open(os.path.join(os.path.dirname(__file__), fname)) as f:
        return f.read()

setup(
    name="biasaway",
    description="a tool to generate nucleotide composition-matched DNA sequences",
    version=VERSION,
    author="Anthony Mathelier and Aziz Khan",
    license='GPL',
    platforms='linux/unix',
    author_email="anthony.mathelier@ncmm.uio.no",
    url="https://bitbucket.org/CBGR/biasaway/src/master/",
    long_description=readme("README.rst"),
    long_description_content_type='text/markdown',
    package_dir={'biasaway': 'biasaway'},

    packages=['biasaway',
        #'biasaway.example_data'
        ],

    scripts=['biasaway/biasaway',
                   ],
    include_package_data=True,
    install_requires = install_requires,
    classifiers=CLASSIFIERS,
)



"""
Module matching %GC compo distribution b/w fg and bg w/i a sliding win.

Modified by Aziz Khan on October 29, 2019
Modified by A. Mathelier in March 2020

"""


from __future__ import print_function
import sys
import random
import numpy
from biasaway.utils import GC, dinuc_count, open_for_parsing
from Bio import SeqIO
from biasaway.GC_compo_matching import print_in_bg_dir, get_bins_from_bg_dir
from biasaway.GC_compo_matching import get_bins_len_from_bg_dir


def GC_info(seq, win_len, step):
    """
    Calculate needed %GC information.

    Calculate G+C content, minimal %GC in sliding windows, maximal %GC in
    sliding windows, stdev of %GC in sliding windows, and CV of %GC in sliding
    windows
    For GC content, it returns the percentage (float between 0 and 100).
    Copes mixed case sequences, and with the ambiguous nucleotide S (G or C)
    when counting the G and C content.    The percentage is calculated against
    the length of the sequence using A,C,G,T,S,W with Ns, e.g.:
    >>> GC("ACTGN")
    50.0
    Note that this will return zero for an empty sequence.

    """

    gc = GC(seq)
    tmp_gc = []
    if win_len >= len(seq):
        return gc, gc, gc, 0, 0
    for i in range(0, len(seq) - win_len):
        tmp_gc.append(GC(seq[i:i + win_len]))
    sd = numpy.std(tmp_gc)
    # Applying +1 to GC to make sure we do not divide by 0

    return gc, min(tmp_gc), max(tmp_gc), sd, 100. * sd / (gc+1.)


def avg_and_sd_gc_info(gc_info):
    """
    Compute information needed w/i the windows.

    Compute averages and standard deviations of all the information within
    each one of the gc bins contained in gc_info.
    Return the dictionary storing info in %GC bins.

    """
    gc_bins = [[0]] * 101
    for gc in range(0, 101):
        if gc_info[gc]:
            min_gc = [x[0] for x in gc_info[gc]]
            max_gc = [x[1] for x in gc_info[gc]]
            sd_gc = [x[2] for x in gc_info[gc]]
            cv_gc = [x[3] for x in gc_info[gc]]
            min_gc_avg = numpy.average(min_gc)
            min_gc_sd = numpy.std(min_gc)
            max_gc_avg = numpy.average(max_gc)
            max_gc_sd = numpy.std(max_gc)
            sd_gc_avg = numpy.average(sd_gc)
            sd_gc_sd = numpy.std(sd_gc)
            cv_gc_avg = numpy.average(cv_gc)
            cv_gc_sd = numpy.std(cv_gc)
            gc_bins[gc] = [(len([x[0] for x in gc_info[gc]]),
                            (min_gc_avg, min_gc_sd), (max_gc_avg, max_gc_sd),
                            (sd_gc_avg, sd_gc_sd), (cv_gc_avg, cv_gc_sd))]
    return gc_bins


def fg_GC_bins(fg, winlen, step):
    """
    Get %GC info for foreground sequences.

    Compute G+C content for all sequences in the foreground and store the
    information in a list. To each G+C percentage bin, we associate the number
    of sequences falling in the corresponding bin

    Return the corresponding lists.

    """

    with open_for_parsing(fg) as stream:
        tmp_gc_bins = []
        gc_list = []
        lengths = []
        dinuc = [0] * 16
        for _ in range(0, 101):
            tmp_gc_bins.append([])
        for record in SeqIO.parse(stream, "fasta"):
            gc, min_gc, max_gc, sd_gc, cv_gc = GC_info(record.seq, winlen,
                                                       step)
            gc_list.append(gc)
            # python 3 fix
            gc = int(round(gc))
            tmp_gc_bins[gc].append((min_gc, max_gc, sd_gc, cv_gc))
            lengths.append(len(record.seq))
            dinuc = [x + y for x, y in zip(dinuc, dinuc_count(record.seq))]
    return gc_list, avg_and_sd_gc_info(tmp_gc_bins), lengths, dinuc


def avg_and_sd_len_gc_info(l_dic, gc_info):
    """
    Get needed info about lengths.

    Compute averages and standard deviations of all the information within each
    one of the gc bins contained in gc_info.

    Return the info in %GC bins.

    """

    gc_bins = [[0]] * 101
    for gc in range(0, 101):
        if gc_info[gc]:
            min_gc = [x[0] for x in gc_info[gc]]
            max_gc = [x[1] for x in gc_info[gc]]
            sd_gc = [x[2] for x in gc_info[gc]]
            cv_gc = [x[3] for x in gc_info[gc]]
            min_gc_avg = numpy.average(min_gc)
            min_gc_sd = numpy.std(min_gc)
            max_gc_avg = numpy.average(max_gc)
            max_gc_sd = numpy.std(max_gc)
            sd_gc_avg = numpy.average(sd_gc)
            sd_gc_sd = numpy.std(sd_gc)
            cv_gc_avg = numpy.average(cv_gc)
            cv_gc_sd = numpy.std(cv_gc)
            gc_bins[gc] = [(l_dic[gc], (min_gc_avg, min_gc_sd), (max_gc_avg,
                                                                 max_gc_sd),
                            (sd_gc_avg, sd_gc_sd), (cv_gc_avg, cv_gc_sd))]
    return gc_bins


def fg_len_GC_bins(fg, winlen, step):
    """
    Get needed lengths info for foreground sequences.

    Compute G+C content for all sequences in the foreground and store the
    information in a list. To each G+C percentage bin, we associate the number
    of sequences falling in the corresponding bin.

    Return the corresponding info in lists.

    """

    with open_for_parsing(fg) as stream:
        tmp_gc_bins = []
        gc_list = []
        lengths = []
        dinuc = [0] * 16
        l_dic = []
        for _ in range(0, 101):
            tmp_gc_bins.append([])
            l_dic.append({})
        for record in SeqIO.parse(stream, "fasta"):
            gc, min_gc, max_gc, sd_gc, cv_gc = GC_info(record.seq, winlen,
                                                       step)
            gc_list.append(gc)
            tmp_gc_bins[gc].append((min_gc, max_gc, sd_gc, cv_gc))
            length = len(record)
            if length in l_dic[gc]:
                l_dic[gc][length] += 1
            else:
                l_dic[gc][length] = 1
            lengths.append(length)
            dinuc = [x + y for x, y in zip(dinuc, dinuc_count(record.seq))]
    return gc_list, avg_and_sd_len_gc_info(l_dic, tmp_gc_bins), lengths, dinuc


def bg_GC_bins(bg, bg_dir):
    """
    Get %GC info for background sequences.

    Compute G+C content for all sequences in the background and store the
    information in a list. To each G+C percentage bin, we associate the
    corresponding sequence names with information about GC composition within
    sliding windows.

    Return info in lists.

    """

    with open_for_parsing(bg) as stream:
        gc_bins = []
        gc_list = []
        lengths = []
        dinuc = [0] * 16
        for _ in range(0, 101):
            gc_bins.append([])
        for record in SeqIO.parse(stream, "fasta"):
            gc = GC(record.seq)
            gc_list.append(gc)
            gc_bins[gc].append(record)
            lengths.append(len(record.seq))
            dinuc = [x + y for x, y in zip(dinuc, dinuc_count(record.seq))]
    print_in_bg_dir(gc_bins, bg_dir)
    return gc_list, gc_bins, lengths, dinuc


def bg_len_GC_bins(bg, bg_dir):
    """
    Get lengths info for background sequences.

    Compute G+C content for all sequences in the background and store the
    information in a list. To each G+C percentage bin, we associate the
    corresponding sequence names with information about GC composition within
    sliding windows.

    Return info in lists.

    """

    with open_for_parsing(bg) as stream:
        gc_bins = []
        gc_list = []
        lengths = []
        dinuc = [0] * 16
        for _ in range(0, 101):
            gc_bins.append({})
        for record in SeqIO.parse(stream, "fasta"):
            gc = GC(record.seq)
            gc_list.append(gc)
            if len(record) in gc_bins[gc]:
                gc_bins[gc][len(record)].append(record)
            else:
                gc_bins[gc][len(record)] = [record]
            lengths.append(len(record.seq))
            dinuc = [x + y for x, y in zip(dinuc, dinuc_count(record.seq))]
    print_in_bg_dir(gc_bins, bg_dir, True)
    return gc_list, gc_bins, lengths, dinuc


def inside(val, center, stdev, deviation):
    """ Return if the value is inside the asked range. """

    return (val >= center - deviation * stdev and
            val <= center + deviation * stdev)


def same_bg(min_gc, max_gc, sd_gc, cv_gc, fg, deviation):
    """ Return if the background seq info are matching the fg seq info. """

    (_, (min_gc_avg, min_gc_sd), (max_gc_avg, max_gc_sd), (sd_gc_avg,
                                                           sd_gc_sd),
     (cv_gc_avg, cv_gc_sd)) = fg[0]
    return (inside(min_gc, min_gc_avg, min_gc_sd, deviation) and
            inside(max_gc, max_gc_avg, max_gc_sd, deviation) and
            inside(sd_gc, sd_gc_avg, sd_gc_sd, deviation) and
            inside(cv_gc, cv_gc_avg, cv_gc_sd, deviation))


def extract_random_sample(bg, fg, nb, deviation, winlen, step):
    """ Return the # of samples found and the samples. """

    random.seed()
    random.shuffle(bg)
    index = 0
    sample = []
    while index < len(bg) and nb:
        record = bg[index]
        _, min_gc, max_gc, sd_gc, cv_gc = GC_info(record.seq, winlen, step)
        if same_bg(min_gc, max_gc, sd_gc, cv_gc, fg, deviation):
            sample.append(record)
            nb -= 1
        index += 1
    return nb, sample


def generate_sequences(fg_bins, bg_bins, bg_dir, deviation, winlen, step,
                       nfold):
    """
    Choose randomly the background sequences in each bin of GC%.

    The same distribution as the one of foreground sequences with a nfold ratio
    is asked.

    Return the sequences with their lengths.

    """

    gc_list = []
    lengths = []
    dinuc = [0] * 16
    for percent in range(0, 101):
        if fg_bins[percent][0]:
            nb = fg_bins[percent][0][0] * nfold
            if bg_bins:
                bin_seq = bg_bins[percent]
            else:
                bin_seq = get_bins_from_bg_dir(bg_dir, percent)
            left, sample = extract_random_sample(bin_seq, fg_bins[percent], nb,
                                                 deviation, winlen, step)
            if left:
                sys.stderr.write("""\n*** WARNING ***
                Sample larger than population for {0:d}% G+C content:
                {1:d} needed and {2:d} obtained\n""".format(percent, nb, nb -
                                                            left))
                gc_list.extend([percent] * (nb - left))
            else:
                gc_list.extend([percent] * nb)
            for r in sample:
                print(r.format("fasta"), end='')
                lengths.append(len(r.seq))
                dinuc = [x + y for x, y in zip(dinuc, dinuc_count(r.seq))]
    return gc_list, lengths, dinuc


def extract_seq_rec(size, nb, bg_keys, bg, accu, index, fg, deviation, winlen,
                    step):
    """
    Extract "nb" sequences w/ sizes equal, as close as possible, to "size" nt.

    This is a tail recursive function with the accumulator "accu" looking for
    sizes "bg_keys" in the bg set "bg".

    Return the info for a recursive function.

    """

    # There could be too many recursions if a lot of the background sequences
    # have the same length. Hence, we set a max number of recursion to avoid
    # going too deep when it does not make sense.
    sys.setrecursionlimit(10000)
    random.seed()
    if not (bg_keys and nb):
        return accu, nb, bg_keys
    if index > len(bg_keys) - 1:
        return extract_seq_rec(size, nb, bg_keys, bg, accu, index - 1, fg,
                               deviation, winlen, step)
    if not bg_keys:
        if bg[bg_keys[index - 1]]:
            random.shuffle(bg[bg_keys[index - 1]])
            record = bg[bg_keys[index - 1]][0]
            _, min_gc, max_gc, sd_gc, cv_gc = GC_info(record.seq, winlen, step)
            if same_bg(min_gc, max_gc, sd_gc, cv_gc, fg, deviation):
                accu.append(record)
                bg[bg_keys[index - 1]] = bg[bg_keys[index - 1]][1:]
                return extract_seq_rec(size, nb - 1, bg_keys, bg, accu, index,
                                       fg, deviation, winlen, step)
            else:
                return extract_seq_rec(size, nb, bg_keys, bg, accu, index, fg,
                                       deviation, winlen, step)
        else:
            return accu, nb, bg_keys
    if bg_keys[index] >= size:
        if (index == 0 or not bg[bg_keys[index - 1]] or
                bg_keys[index] - size < size - bg_keys[index - 1]):
            random.shuffle(bg[bg_keys[index]])
            record = bg[bg_keys[index]][0]
            if bg[bg_keys[index]][1:]:
                bg[bg_keys[index]] = bg[bg_keys[index]][1:]
            else:
                bg[bg_keys[index]] = bg[bg_keys[index]][1:]
                del bg_keys[index]
            _, min_gc, max_gc, sd_gc, cv_gc = GC_info(record.seq, winlen, step)
            if same_bg(min_gc, max_gc, sd_gc, cv_gc, fg, deviation):
                accu.append(record)
                return extract_seq_rec(size, nb - 1, bg_keys, bg, accu, index,
                                       fg, deviation, winlen, step)
            else:
                return extract_seq_rec(size, nb, bg_keys, bg, accu, index, fg,
                                       deviation, winlen, step)
        else:
            random.shuffle(bg[bg_keys[index - 1]])
            record = bg[bg_keys[index - 1]][0]
            if bg[bg_keys[index - 1]][1:]:
                bg[bg_keys[index - 1]] = bg[bg_keys[index - 1]][1:]
            else:
                bg[bg_keys[index - 1]] = bg[bg_keys[index - 1]][1:]
                del bg_keys[index - 1]
                index = index - 1
            _, min_gc, max_gc, sd_gc, cv_gc = GC_info(record.seq, winlen, step)
            if same_bg(min_gc, max_gc, sd_gc, cv_gc, fg, deviation):
                accu.append(record)
                return extract_seq_rec(size, nb - 1, bg_keys, bg, accu, index,
                                       fg, deviation, winlen, step)
            else:
                return extract_seq_rec(size, nb, bg_keys, bg, accu, index, fg,
                                       deviation, winlen, step)
    elif index == len(bg_keys) - 1:
        random.shuffle(bg[bg_keys[index]])
        record = bg[bg_keys[index]][0]
        if bg[bg_keys[index]][1:]:
            bg[bg_keys[index]] = bg[bg_keys[index]][1:]
        else:
            bg[bg_keys[index]] = bg[bg_keys[index]][1:]
            del bg_keys[index]
            index = index - 1
        _, min_gc, max_gc, sd_gc, cv_gc = GC_info(record.seq, winlen, step)
        if same_bg(min_gc, max_gc, sd_gc, cv_gc, fg, deviation):
            accu.append(record)
            return extract_seq_rec(size, nb - 1, bg_keys, bg, accu, index, fg,
                                   deviation, winlen, step)
        else:
            return extract_seq_rec(size, nb, bg_keys, bg, accu, index, fg,
                                   deviation, winlen, step)
    else:
        return extract_seq_rec(size, nb, bg_keys, bg, accu, index + 1, fg,
                               deviation, winlen, step)


def generate_len_sequences(fg_bins, bg_bins, bg_dir, deviation, winlen, step,
                           nfold):
    """
    Choose randomly the background sequences in each bin of GC%.

    with the same distribution as the one of foreground sequences with a nfold
    ratio.

    Return the sequences and their lengths.

    """

    sys.setrecursionlimit(10000)
    gc_list = []
    lengths = []
    dinuc = [0] * 16
    for percent in range(0, 101):
        if fg_bins[percent][0]:
            nb = sum(fg_bins[percent][0][0].values()) * nfold
            if bg_bins:
                bin_seq = bg_bins[percent]
            else:
                bin_seq = get_bins_len_from_bg_dir(bg_dir, percent)
            sequences = []
            bg_keys = sorted(bin_seq.keys())
            for size in fg_bins[percent][0][0].keys():
                nb_to_retrieve = fg_bins[percent][0][0][size] * nfold
                seqs, _, bg_keys = extract_seq_rec(size, nb_to_retrieve,
                                                   bg_keys, bin_seq, [], 0,
                                                   fg_bins[percent], deviation,
                                                   winlen, step)
                sequences.extend(seqs)
            nb_match = len(sequences)
            if nb_match != nb:
                sys.stderr.write("""\n*** WARNING ***
                Sample larger than population for {0:d}% G+C content:
                {1:d} needed and {2:d} obtained\n""".format(percent, nb,
                                                            nb_match))
            gc_list.extend([percent] * (nb_match))
            for r in sequences:
                print("{0:s}".format(r.format("fasta")), end='')
                lengths.append(len(r))
                dinuc = [x + y for x, y in zip(dinuc, dinuc_count(r.seq))]
    return gc_list, lengths, dinuc



"""
Modified by Aziz Khan on October 29, 2019
Modified by A. Mathelier in March 2020
"""

from Bio import SeqIO
from os.path import splitext
import gzip
import re


def open_for_parsing(filename):
    """
    Open the file with the given filename for parsing. Where the file appears
    to be a gzip-compressed file it is opened in the appropriate way to provide
    access to the decompressed text.
    """
    ext = splitext(filename)[-1]
    if ext == ".gz":
        return gzip.open(filename, "rt")  # text flag "t" needed in Python 3
    else:
        return open(filename)


def GC(seq):
    """
    Calculate G+C content, return the percentage (float between 0 and 100).
    Copes mixed case sequences, and with the ambiguous nucleotide S (G or C)
    when counting the G and C content. The percentage is calculated against
    the length of the sequence using A,C,G,T,S,W with Ns, e.g.:
    >>> GC("ACTGN")
    50.0 Note that this will return zero for an empty sequence.

    """
    try:
        gc = sum(map(seq.count, ['G', 'C', 'g', 'c', 'S', 's']))
        length = sum(map(seq.count, ['G', 'C', 'A', 'T', 'S', 'W', 'g', 'c',
                                     'a', 't', 's', 'w']))
        # fix for py3 issues
        return int(round(gc * 100 / length))
    except ZeroDivisionError:
        return 0


def dinuc_count(seq):
    """
    Calculate dinucleotide composition of a sequence.
    Only considers A, C, G, T
    """
    dinuc = ['AA', 'AT', 'AG', 'AC', 'TA', 'TT', 'TG', 'TC', 'GA', 'GT', 'GG',
             'GC', 'CA', 'CT', 'CG', 'CC']
    return map(seq.count, dinuc)


def get_seqs(f):
    seqs = []
    fg_gc_list = []
    fg_lengths = []
    dinuc = [0] * 16
    with open_for_parsing(f) as stream:
        for record in SeqIO.parse(stream, "fasta"):
            record.seq = record.seq.upper()
            seqs.append(record)
            fg_gc_list.append(GC(record.seq))
            fg_lengths.append(len(record.seq))
            dinuc = [x + y for x, y in zip(dinuc, dinuc_count(record.seq))]
    return seqs, fg_gc_list, fg_lengths, dinuc


def init_compo(length):
    dico = []
    for i in range(1, length):
        dico.append({})
        j = i - 1
        dico[j]["AA"] = 0.0
        dico[j]["AC"] = 0.0
        dico[j]["AT"] = 0.0
        dico[j]["AG"] = 0.0
        dico[j]["CA"] = 0.0
        dico[j]["CC"] = 0.0
        dico[j]["CT"] = 0.0
        dico[j]["CG"] = 0.0
        dico[j]["GA"] = 0.0
        dico[j]["GC"] = 0.0
        dico[j]["GG"] = 0.0
        dico[j]["GT"] = 0.0
        dico[j]["TA"] = 0.0
        dico[j]["TC"] = 0.0
        dico[j]["TG"] = 0.0
        dico[j]["TT"] = 0.0
    return dico


def length(record):
    return len(record.seq)


def all_pos_dinuc(compo, max_length):
    distrib = {}
    composition = {}
    for key in compo[0].keys():
        cpt = 0.0
        for i in range(1, max_length):
            cpt += compo[i - 1][key]
        composition[key] = cpt
    for k in composition.keys():
        cpt = 4.0  # WARNING: WE DO NOT TAKE ANY 'N' INTO ACCOUNT
        first = k[0]
        for key in composition.keys():
            if key[0] == first:
                cpt += composition[key]
        distrib[k] = (composition[k] + 1.0) / cpt
    return distrib


def pos_by_pos_dinuc(compo, seq_length):
    distrib = init_compo(seq_length)
    for j in range(1, seq_length):
        for k in compo[j - 1].keys():
            if re.search("N", k):
                distrib[j - 1][k] = 0.0
            else:
                cpt = 4.0
                first = k[0]
                for key in compo[j - 1].keys():
                    if key[0] == first:
                        cpt += compo[j - 1][key]
                distrib[j - 1][k] = (compo[j - 1][k] + 1.0) / cpt
    return distrib


def compute_dinuc_distrib(seqs, b=False):
    max_length = max(map(length, seqs))
    compo = init_compo(max_length)
    for i in range(0, len(seqs)):
        seq_length = len(seqs[i].seq)
        for j in range(1, seq_length):
            if seqs[i].seq[j - 1] != 'N' and seqs[i].seq[j] != 'N':
                compo[j - 1]["%s" % (seqs[i].seq[(j - 1):(j + 1)])] += 1.0

    if b:  # Dinucleotide distrib over all positions
        return all_pos_dinuc(compo, max_length)
    else:  # Dinucleotide distrib position by position
        return pos_by_pos_dinuc(compo, seq_length)


def print_dinuc_distrib(dinuc, output):
    import sys
    stream = sys.stdout
    if output:
        stream = open(output, "w")
    for j in range(0, len(dinuc)):
        stream.write("%f, %f, %f, %f, %f, %f, %f, %f, %f," % (dinuc[j]["AA"],
                                                              dinuc[j]["AC"],
                                                              dinuc[j]["AG"],
                                                              dinuc[j]["AT"],
                                                              dinuc[j]["AN"],
                                                              dinuc[j]["CA"],
                                                              dinuc[j]["CC"],
                                                              dinuc[j]["CG"],
                                                              dinuc[j]["CT"]))
        stream.write(" %f, %f, %f, %f, %f, %f, %f, %f, %f," % (dinuc[j]["CN"],
                                                               dinuc[j]["GA"],
                                                               dinuc[j]["GC"],
                                                               dinuc[j]["GG"],
                                                               dinuc[j]["GT"],
                                                               dinuc[j]["GN"],
                                                               dinuc[j]["TA"],
                                                               dinuc[j]["TC"],
                                                               dinuc[j]["TG"]))
        stream.write(" %f, %f, %f, %f, %f, %f, %f\n" % (dinuc[j]["TT"],
                                                        dinuc[j]["TN"],
                                                        dinuc[j]["NA"],
                                                        dinuc[j]["NC"],
                                                        dinuc[j]["NG"],
                                                        dinuc[j]["NT"],
                                                        dinuc[j]["NN"]))
    stream.close()


def compute_nt_distrib(seqs):
    cpt = 4.0
    distrib = {}
    for l in "ACGT":
        distrib[l] = 1.0
    for seq in seqs:
        for letter in seq:
            if letter != 'N':
                distrib[letter] += 1.0
                cpt += 1.0
    for letter in "ACGT":
        distrib[letter] /= cpt
    return distrib


def split_seq(seq):
    return re.split('([!ACGT]+)', seq)


def single_value(the_array):
    import numpy as np
    return len(np.unique(the_array)) == 1


def make_gc_plot(fg_gc, bg_gc, plot_filename):
    """
    Compute the density GC composition plots for background and input.
    """
    import matplotlib
    matplotlib.use('Agg')
    import seaborn as sns
    import matplotlib.pyplot as plt
    plot_hist = False
    plot_kde = True
    ylab = "density"
    # One cannot compute kde if there is a single value in the array
    if single_value(fg_gc) or single_value(bg_gc):
        plot_hist = True
        plot_kde = False
        ylab = "frequency"
    plt.figure()
    plot = sns.distplot(fg_gc, hist=plot_hist, kde=plot_kde,
                        kde_kws={'shade': True, 'linewidth': 3},
                        label='input')
    plot = sns.distplot(bg_gc, hist=plot_hist, kde=plot_kde,
                        kde_kws={'shade': True, 'linewidth': 3},
                        label='generated')
    plt.legend()
    plot.set(xlabel="%GC", ylabel=ylab)
    plt.savefig("{0}_gc_plot.png".format(plot_filename))


def make_len_plot(fg_len, bg_len, plot_filename):
    """
    Compute the density length plot for the background, the input and the
    matching background datasets.
    """
    import matplotlib
    matplotlib.use('Agg')
    import seaborn as sns
    import matplotlib.pyplot as plt
    plot_hist = False
    plot_kde = True
    ylab = "density"
    # One cannot compute kde if there is a single value in the array
    if single_value(fg_len) or single_value(bg_len):
        plot_hist = True
        plot_kde = False
        ylab = "frequency"
    plt.figure()
    plot = sns.distplot(fg_len, hist=plot_hist, kde=plot_kde,
                        kde_kws={'shade': True, 'linewidth': 3},
                        label='input')
    plot = sns.distplot(bg_len, hist=plot_hist, kde=plot_kde,
                        kde_kws={'shade': True, 'linewidth': 3},
                        label='generated')
    plt.legend()
    plot.set(xlabel="length", ylabel=ylab)
    plot.get_figure().savefig("{0}_length_plot.png".format(plot_filename))


def make_dinuc_plot(fg_dinuc, bg_dinuc, plot_filename):
    """
    Plot the dinucleotide composition of input and background sequences.
    """
    import pandas as pd
    import matplotlib
    matplotlib.use('Agg')
    import seaborn as sns
    import matplotlib.pyplot as plt
    nuc = ['A', 'C', 'G', 'T']
    fg_total = sum(fg_dinuc)
    fg_dinuc = [val / fg_total for val in fg_dinuc]
    fg_df = pd.DataFrame({'A': fg_dinuc[0:4], 'C': fg_dinuc[4:8],
                          'G': fg_dinuc[8:12], 'T': fg_dinuc[12:16]},
                         index=nuc)
    bg_total = sum(bg_dinuc)
    bg_dinuc = [val / bg_total for val in bg_dinuc]
    bg_df = pd.DataFrame({'A': bg_dinuc[0:4], 'C': bg_dinuc[4:8],
                          'G': bg_dinuc[8:12], 'T': bg_dinuc[12:16]},
                         index=nuc)
    mini = min(min(fg_dinuc), min(bg_dinuc))
    maxi = max(max(fg_dinuc), max(bg_dinuc))
    fig, (ax1, ax2) = plt.subplots(ncols=2)
    fig.subplots_adjust(wspace=0.05)
    sns.heatmap(fg_df, cmap="icefire", ax=ax1, cbar=False, annot=True,
                fmt='.2f', vmin=mini, vmax=maxi)
    ax1.set(xlabel="first nucleotide", ylabel="second nucleotide",
            title="input")
    fig.colorbar(ax1.collections[0], ax=ax1, location="left",
                 use_gridspec=False, pad=0.2)
    sns.heatmap(bg_df, cmap="icefire", ax=ax2, cbar=False, annot=True,
                fmt='.2f', vmin=mini, vmax=maxi)
    ax2.set(xlabel="first nucleotide", title="generated")
    fig.colorbar(ax2.collections[0], ax=ax2, location="right",
                 use_gridspec=False, pad=0.2)
    ax2.yaxis.tick_right()
    ax2.tick_params(labelrotation=0)
    plt.savefig("{0}_dinuc_plot.png".format(plot_filename))
